"""
coding: utf-8
@Software: PyCharm
@Time:  0:41
@Author: Fake77
@Module Name:
"""
import sys
from TB4hooky.Utils.LinkDealer import LinksDealer
import os
from loguru import logger
from aiohttp import ClientSession
import asyncio
import aiohttp
from tqdm import tqdm


class FileHandle(object):
    """
    package = FileHandle("http://127.0.0.1:9000", 'torch')
    package.start_remote_import()

    """

    def __init__(self, _endpoint: str, package: str, loop):
        # 判断是否为dist-info 包文件
        if package.endswith("dist-info"):
            package = package
        else:
            package = package.replace(".", '/')
        # 生成根路径
        self._root_path = _endpoint
        # 包路径
        self._endpoint = _endpoint + '/' + package
        self._package = package
        # 初始化链接库
        self._links = LinksDealer.get_filehandle_links(_endpoint + '/' + package)
        # 获取当前安装目录
        self._install_path = FileHandle.get_install_path()
        # 如果获取不到安装目录则在当前文件夹下创建安装目录
        if self._install_path is None:
            self._install_path = '\\site-packages'
        # 下载文件列表
        self._files = []
        # 事件循环
        self._event_loop = loop
        # 设置并发信号量
        self.semaphore = asyncio.Semaphore(100)
        # 设置重发列表
        self._retry = []

    @staticmethod
    def get_install_path():
        """
        获取安装路径 一般为 site-packages
        """
        for path in sys.path:
            if path.split('\\')[-1] == 'site-packages':
                return path

    @staticmethod
    def is_directory(path):
        """
        判断是否为文件
        directory/ 目录 -> True
        hello.txt 文件 -> False
        """
        if path[-1] == '/':
            return True
        return False

    async def download_file(self, path: str, pbar: tqdm):
        """
        下载文件
        :param: path: url
        :param: pbar: tqdm object
        """
        async with self.semaphore:
            try:
                async with ClientSession() as session:
                    async with session.get(path) as content:
                        filepath = self._install_path + path[len(self._root_path):].replace("/", '\\')

                        directory_path = "\\".join(filepath.split('\\')[:-1])

                        if not os.path.exists(directory_path):
                            os.makedirs(directory_path)

                        with open(filepath, 'wb') as f:
                            f.write(await content.read())
                        pbar.update(1)
            except aiohttp.client_exceptions.ClientOSError:
                self._retry.append(path)

    def get_all_file(self, endpoint: str):
        """
        遍历当前endpoint目录下的全部文件
        """
        # 递归获取该包的全部文件
        for sub_path in LinksDealer.get_filehandle_links(endpoint):
            if endpoint[-1] != '/':
                path = endpoint + '/' + sub_path
            else:
                path = endpoint + sub_path
            if self.is_directory(path):
                self.get_all_file(path)
            else:
                self._files.append(path)

    async def main(self, link_lst, pbar):
        tasks = []
        for url in link_lst:
            tasks.append(asyncio.create_task(self.download_file(url, pbar)))
        await asyncio.wait(tasks)

    def start_remote_import(self):
        # 提交远程文件路径到self._files
        self.get_all_file(self._endpoint)

        # 判断远程包是否存在
        if len(self._files) == 0:
            raise ImportError(f"Not such Package {self._package} in {self._endpoint}.")

        logger.info(f"remote import {self._package} from {self._endpoint} total file: {len(self._files)}")

        # fetch 远程包
        with tqdm(total=len(self._files)) as pbar:
            self._event_loop.run_until_complete(self.main(self._files, pbar))
            if len(self._retry) != 0:
                self._event_loop.run_until_complete(self.main(self._retry, pbar))




"""
coding: utf-8
@Software: PyCharm
@Time:  6:23
@Author: Fake77
@Module Name:
"""
import base64


class CacheMinx:
    __slots__ = ()

    def __setitem__(self, key, value):
        if key in ("co_lnotab", "co_codestring"):
            value = base64.b64decode(value.encode())
        if key in ("co_argcount", "co_kwonlyargcount", "co_nlocals",
                   "co_stacksize", "co_flags", "co_firstlineno"):
            value = int(value)
        if key in ("co_consts", "co_names", "co_varnames", "co_freevars"):
            value = tuple(value)
        return super().__setitem__(key, value)


class Cache(CacheMinx, dict):
    ...



"""
coding: utf-8
@Software: PyCharm
@Time:  0:32
@Author: Fake77
@Module Name:
"""
from TB4hooky.Utils.ImportUtil import FileHandle
from importlib import import_module
import time


class AddressInjection(object):
    ADDRESS = None

    @staticmethod
    def register_backup_imp_server(_f):
        AddressInjection.ADDRESS = _f

    def __call__(self, func):

        def arg(*args, **kwargs):
            result = func(*args, **kwargs, remote_addr=self.ADDRESS)
            return result

        return arg


@AddressInjection()
def handle_exception(excType, excValue, tb, remote_addr):
    if excType is ModuleNotFoundError or ImportError:
        module_name = str(excValue).split("'")[-2]
        try:
            FileHandle(remote_addr, module_name).start_remote_import()
            time.sleep(0.5)
            globals()[module_name] = import_module(module_name)

        except Exception as e:
            print(e)
            return

    else:
        raise excType(excValue, tb)



"""
coding: utf-8
@Software: PyCharm
@Time:  7:06
@Author: Fake77
@Module Name:
"""
import sys


class GlobalException(object):

    def __init__(self, exc_handle):
        sys.excepthook = exc_handle

    @classmethod
    def register_global_exception(cls, exc_handle):
        cls(exc_handle)




import asyncio
import importlib
import time
from loguru import logger
from TB4hooky.netLoader.MetaFinder import RemoteMetaFinder
from TB4hooky.netLoader.SourceCodeManager import SourceCodeManager
from TB4hooky.netLoader.ComplexImport import FileHandle
from TB4hooky.Utils.LinkDealer import LinksDealer

import sys

_meta = {}
_scm = SourceCodeManager()
_endpoint = ""


def register(endpoint: str) -> None:
    global _endpoint
    _endpoint = endpoint
    if not hasattr(_meta, endpoint):
        finder = RemoteMetaFinder(endpoint, _scm)
        _meta[endpoint] = finder
        sys.meta_path.append(finder)

def remote_import(_pkg: str):
    if _endpoint:
        loop = asyncio.get_event_loop()
        links = LinksDealer.get_link(_endpoint)
        for link in links:
            if link.startswith(_pkg.split('.')[0]):
                logger.success(f"Find pkg {link.replace('/', '')}!")
                FileHandle(_endpoint, link, loop).start_remote_import()
                time.sleep(0.2)
        print(FileHandle.get_install_path())
        sys.path.append(FileHandle.get_install_path() + '\\' + _pkg.split('.')[0])
        pkg = importlib.import_module(_pkg)
        return pkg


def unload(endpoint: str) -> None:
    if hasattr(_meta, endpoint):
        finder = _meta.pop(endpoint)
        sys.meta_path.remove(finder)



"""
coding: utf-8
@Software: PyCharm
@Time:  3:32
@Author: Fake77
@Module Name:
"""
from TB4hooky.hooky.Hooker import CodeHooker

__all__ = [
    'REMOTE_FUNCTION_MAPPING',
    'ServerSerialize',
]

REMOTE_FUNCTION_MAPPING = {}


class ServerSerialize(object):
    def __init__(self, count):
        self.count = count

    def __call__(self, func):
        def recv_args(*args, **kwargs):
            # func callable校验
            assert callable(func)
            func_serialize = CodeHooker.serialize_init(func)
            serialize_result = func_serialize.serialize_func(self.count)
            result = func(*args, **kwargs)
            REMOTE_FUNCTION_MAPPING[func.__name__] = serialize_result
            return result
        return recv_args

